/*
############################################################################
# This file contains the word functions for the pfth Forth interpreter.
#
# Copyright (c) 2012 Dave Hein
# MIT Licensed
############################################################################
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#include "pfth.h"
#include "words.h"

char pad[256];
char *printptr = pad + 255;

static int  ConvertSignedNumber(int value);
static int  ConvertUnsignedNumber(unsigned int value);

int basefunc(void *parm)
{
    push((int)&base);
    return errorcode;
}

int FAST fetchfunc(void *parm)
{
    if (stkidx < 1)
      return RETURN_ERROR;
    stack[stkidx-1] = *(int *)stack[stkidx-1];
    return errorcode;
}

int FAST storefunc(void *parm)
{
    int *ptr;
    if (stkidx < 2)
      return RETURN_ERROR;
    ptr = (int *)stack[--stkidx];
    *ptr = stack[--stkidx];
    return errorcode;
}

int FAST cfetchfunc(void *parm)
{
    if (stkidx < 1)
      return RETURN_ERROR;
    stack[stkidx-1] = *(unsigned char *)stack[stkidx-1];
    return errorcode;
}

int FAST cstorefunc(void *parm)
{
    unsigned char *ptr;
    if (stkidx < 2)
      return RETURN_ERROR;
    ptr = (unsigned char *)stack[--stkidx];
    *ptr = stack[--stkidx];
    return errorcode;
}

int herefunc(void *parm)
{
    push((int)here);
    return errorcode;
}

int FAST plusfunc(void *parm)
{
    if (stkidx < 2) return RETURN_ERROR;
    stkidx--;
    stack[stkidx-1] = stack[stkidx-1] + stack[stkidx];
    return errorcode;
}

int FAST minusfunc(void *parm)
{
    int value = pop();
    push(pop() - value);
    return errorcode;
}

int multfunc(void *parm)
{
    push(pop() * pop());
    return errorcode;
}

int dividefunc(void *parm)
{
    int value = pop();
    push(pop() / value);
    return errorcode;
}

int dotfunc(void *parm)
{
    int len = ConvertSignedNumber(pop());
    Type(printptr+1, len);
    printf(" ");
    return errorcode;
}

int dotrfunc(void *parm)
{
    int value2 = pop();
    int value1 = pop();
    int len = ConvertSignedNumber(value1);

    while (len < value2--) printf(" ");
    Type(printptr+1, len);
    printf(" ");
    return errorcode;
}

int FAST execfunc(void *parm)
{
    int retval = errorcode;
    dictentryT *entry = (dictentryT *)parm;
    control[ctlidx++] = execaddr;
    execaddr = (int *)ALIGN(entry->name + entry->len);
    while ((entry = (dictentryT *)(*execaddr)))
    {
        if ((retval = EXECUTE(entry)) <= RETURN_EXIT)
            break;
        execaddr++;
    }
    execaddr = control[--ctlidx];
    return retval;
}

int FAST varexecfunc(void *parm)
{
    int retval = errorcode;
    dictentryT *entry = (dictentryT *)parm;
    stack[stkidx++] = (int)ALIGN(entry->name + entry->len);
    control[ctlidx++] = execaddr;
    execaddr = entry->does;
    while ((entry = (dictentryT *)(*execaddr)))
    {
        if ((retval = EXECUTE(entry)) <= RETURN_EXIT)
            break;
        execaddr++;
    }
    execaddr = control[--ctlidx];
    return retval;
}

int addrfunc(void *parm)
{
    dictentryT *entry = (dictentryT *)parm;
    void *ptr = ALIGN(entry->name + entry->len);
    push((int)ptr);
    return errorcode;
}

int immediatefunc(void *parm)
{
    if (lastword) lastword->flags |= FLAG_IMMEDIATE;
    return errorcode;
}

int createfunc(void *parm)
{
    int len;
    char *wordptr = Word(&len);
    Create(wordptr, len, addrfunc);
    return errorcode;
}

int allotfunc(void *parm)
{
    here += pop();
    return errorcode;
}

int compileonfunc(void *parm)
{
    compileflag = 1;
    return errorcode;
}

int compileofffunc(void *parm)
{
    compileflag = 0;
    return errorcode;
}

void MakeEntry(char *name, wordfuncP wordfunc, int flags)
{
    Create(name, strlen(name), wordfunc);
    lastword->flags = flags;
}

int colonfunc(void *parm)
{
    int len;
    char *wordptr = Word(&len);
    Create(wordptr, len, execfunc);
    compileflag = 1;
    return errorcode;
}

int semicolonfunc(void *parm)
{
    compileflag = 0;
    here = ALIGN(here);
    *((int *)here) = 0;
    here += 4;
    return errorcode;
}

int includefunc(void *parm)
{
    int len;
    char fname[100];
    char *wordptr = Word(&len);

    memcpy(fname, wordptr, len);
    fname[len] = 0;

    Include(fname);
    return errorcode;
}

int evaluatefunc(void *parm)
{
    int len = pop();
    char *ptr = (char *)pop();

    instack[inidx++] = inputptr;
    instack[inidx++] = (char *)inputidx;
    instack[inidx++] = (char *)inputlen;

    Evaluate(ptr, len);

    inputlen  = (int)instack[--inidx];
    inputidx  = (int)instack[--inidx];
    inputptr  = instack[--inidx];

    return errorcode;
}

int FAST toretfunc(void *parm)
{
    pushret(pop());
    return errorcode;
}

int FAST fromretfunc(void *parm)
{
    push(popret());
    return errorcode;
}

int FAST rotfunc(void *parm)
{
    int value3 = pop();
    int value2 = pop();
    int value1 = pop();
    push(value2);
    push(value3);
    push(value1);
    return errorcode;
}

int FAST dropfunc(void *parm)
{
    pop();
    return errorcode;
}

int FAST dupfunc(void *parm)
{
    int value = pop();
    push(value);
    push(value);
    return errorcode;
}

int FAST swapfunc(void *parm)
{
    int value1 = pop();
    int value2 = pop();
    push(value1);
    push(value2);
    return errorcode;
}

int equalfunc(void *parm)
{
    int value2 = pop();
    int value1 = pop();
    push(-(value1 == value2));
    return errorcode;
}

int FAST lessfunc(void *parm)
{
    int value1, value2;
    if (stkidx < 2) return RETURN_ERROR;
    value2 = stack[--stkidx];
    value1 = stack[stkidx-1];
    stack[stkidx-1] = -(value1 < value2);
    return errorcode;
}

int greaterfunc(void *parm)
{
    int value2 = pop();
    int value1 = pop();
    push(-(value1 > value2));
    return errorcode;
}

int FAST andfunc(void *parm)
{
    if (stkidx < 2) return RETURN_ERROR;
    stkidx--;
    stack[stkidx-1] = stack[stkidx-1] & stack[stkidx];
    return errorcode;
}

int orfunc(void *parm)
{
    push(pop() | pop());
    return errorcode;
}

int xorfunc(void *parm)
{
    push(pop() ^ pop());
    return errorcode;
}

int lshiftfunc(void *parm)
{
    unsigned int shift = pop();
    unsigned int value = pop();
    push(value << shift);
    return errorcode;
}

int rshiftfunc(void *parm)
{
    unsigned int shift = pop();
    unsigned int value = pop();
    push(value >> shift);
    return errorcode;
}

int depthfunc(void *parm)
{
    push(stkidx);
    return errorcode;
}

int emitfunc(void *parm)
{
    putchar(pop());
    fflush(stdout);
    return errorcode;
}

int postponefunc(void *parm)
{
    if (compileflag)
    {
        dictentryT *wordentry = (dictentryT *)parm;
        execaddr++;
        here = ALIGN(here);
        *((int *)here) = *execaddr;
        here += 4;
        if (wordentry->flags & FLAG_LITERAL)
        {
            *((int *)here) = *execaddr;
            here += 4;
        }
    }
    return errorcode;
}

int beginfunc(void *parm)
{
    if (!compileflag)
        return RETURN_ERROR;
    control[ctlidx++] = (int *)(here - 4);
    return errorcode;
}

int untilfunc(void *parm)
{
    if (compileflag)
    {
        *(int *)here = (int)parm;
        here += 4;
        *(int *)here = (int)control[--ctlidx];
        here += 4;
    }
    else
    {
        if (!pop())
            execaddr = (int *)execaddr[1];
        else
            execaddr++;
    }
    return errorcode;
}

int againfunc(void *parm)
{
    if (compileflag)
    {
        *(int *)here = (int)parm;
        here += 4;
        *(int *)here = (int)control[--ctlidx];
        here += 4;
    }
    else
    {
        execaddr = (int *)execaddr[1];
    }
    return errorcode;
}

int FAST dofunc(void *parm)
{
    if (compileflag)
    {
        *(int *)here = (int)parm;
        control[ctlidx++] = (int *)(here + 1);
        here += 4;
    }
    else
    {
        int index = pop();
        int limit = pop();
        pushret(limit);
        pushret(index);
    }
    return errorcode;
}

int FAST questiondofunc(void *parm)
{
    if (compileflag)
    {
        *(int *)here = (int)parm;
        here += 4;
        control[ctlidx++] = (int *)(here + 2);
        here += 4;
    }
    else
    {
        int index = pop();
        int limit = pop();
        if (index < limit)
        {
            pushret(limit);
            pushret(index);
            execaddr++;
        }
        else
        {
            execaddr = (int *)execaddr[1];
        }
    }
    return errorcode;
}

int leavefunc(void *parm)
{
    if (compileflag)
    {
        int index;
        *(int *)here = (int)parm;
        here += 4;
        for (index = ctlidx++; index > 0; index--)
        {
            if (((int)control[index-1]) & 3)
            {
                control[index] = (int *)(here + 3);
                break;
            }
            control[index] = control[index-1];
        }
        here += 4;
    }
    else
    {
        popret();
        popret();
        execaddr = (int *)execaddr[1];
    }
    return errorcode;
}

int FAST loopfunc(void *parm)
{
    if (compileflag)
    {
        int flag;
        int *ptr;
        *(int *)here = (int)parm;
        here += 4;
        while (1)
        {
            flag = (int)control[--ctlidx];
            ptr = (int *)(flag & ~3);
            flag &= 3;
            if (flag > 1) *ptr = (int)here;
            if (flag != 3) break;
        }
        *(int *)here = (int)ptr;
        here += 4;
    }
    else
    {
#if 0
        int index = popret();
        int limit = popret();
        dictentryT *entry = (dictentryT *)parm;

        if (entry->name[0] == '+')
            index += pop();
        else
            index++;

        if (index < limit)
        {
            execaddr = (int *)execaddr[1];
            pushret(limit);
            pushret(index);
        }
        else
            execaddr++;
#else
        int index, limit;
        
        if (retidx < 2)
            return RETURN_ERROR;

        index = retstk[retidx-1];
        limit = retstk[retidx-2];
        
        if (++index < limit)
        {
            execaddr = (int *)execaddr[1];
            retstk[retidx-1] = index;
        }
        else
        {
            execaddr++;
            retidx -= 2; 
        }       
#endif
    }
    return errorcode;
}

int FAST plusloopfunc(void *parm)
{
    if (compileflag)
    {
        int flag;
        int *ptr;
        *(int *)here = (int)parm;
        here += 4;
        while (1)
        {
            flag = (int)control[--ctlidx];
            ptr = (int *)(flag & ~3);
            flag &= 3;
            if (flag > 1) *ptr = (int)here;
            if (flag != 3) break;
        }
        *(int *)here = (int)ptr;
        here += 4;
    }
    else
    {
        int index = popret();
        int limit = popret();

        index += pop();

        if (index < limit)
        {
            execaddr = (int *)execaddr[1];
            pushret(limit);
            pushret(index);
        }
        else
            execaddr++;
    }
    return errorcode;
}

int FAST whilefunc(void *parm)
{
    if (compileflag)
    {
        *(int *)here = (int)parm;
        here += 4;
        control[ctlidx++] = (int *)here;
        here += 4;
    }
    else
    {
        if (stkidx < 1) return RETURN_ERROR;
        if (stack[--stkidx])
            execaddr++;
        else
            execaddr = (int *)execaddr[1];
    }
    return errorcode;
}

int recursefunc(void *parm)
{
    if (compileflag)
    {
        *(int *)here = (int)lastword;
        here += 4;
        return errorcode;
    }
    return RETURN_ERROR;
}

int FAST repeatfunc(void *parm)
{
    if (compileflag)
    {
        int *ptr = (int *)control[--ctlidx];
        *(int *)here = (int)parm;
        here += 4;
        *(int *)here = (int)control[--ctlidx];
        *ptr = (int)here;
        here += 4;
    }
    else
    {
        execaddr = (int *)execaddr[1];
    }
    return errorcode;
}

int iffunc(void *parm)
{
    if (compileflag)
    {
        here = ALIGN(here);
        *((int *)here) = (int)parm;
        here += 4;
        control[ctlidx++] = (int *)here;
        *((int *)here) = 12345678;
        here += 4;
    }
    else
    {
        if (!pop())
            execaddr = (int *)execaddr[1];
        else
            execaddr++;
    }
    return errorcode;
}

int elsefunc(void *parm)
{
    if (compileflag)
    {
        int *ptr = control[--ctlidx];
        *ptr = (int)(here + 4);
        here = ALIGN(here);
        *((int *)here) = (int)parm;
        here += 4;
        control[ctlidx++] = (int *)here;
        *((int *)here) = 12345678;
        here += 4;
    }
    else
    {
        execaddr = (int *)execaddr[1];
    }
    return errorcode;
}

int thenfunc(void *parm)
{
    if (compileflag)
    {
        int *ptr = control[--ctlidx];
        *ptr = (int)(here - 4);
    }
    return errorcode;
}

int FAST _literalfunc(void *parm)
{
    execaddr++;
    stack[stkidx++] = *execaddr;
    return errorcode;
}

int literalfunc(void *parm)
{
    if (compileflag)
    {
        here = ALIGN(here);
        *((int *)here) = (int)literalword;
        here += 4;
        *((int *)here) = pop();
        here += 4;
        return errorcode;
    }
    return RETURN_ERROR;
}

int doesfunc(void *parm)
{
    if (lastword)
    {
        lastword->func = varexecfunc;
        lastword->does = (int *)(execaddr + 1);
        return RETURN_EXIT;
    }
    return RETURN_ERROR;
}

int byefunc(void *parm)
{
    return RETURN_BYE;
}

int udotfunc(void *parm)
{
    int len = ConvertUnsignedNumber(pop());
    Type(printptr+1, len);
    printf(" ");
    return errorcode;
}

int ulessfunc(void *parm)
{
    unsigned int value2 = pop();
    unsigned int value1 = pop();
    push(-(value1 < value2));
    return errorcode;
}

int typefunc(void *parm)
{
    int num = pop();
    char *ptr = (char *)pop();
    Type(ptr, num);
    return errorcode;
}

int sourcefunc(void *parm)
{
    push((int)inputptr);
    push(inputlen);
    return errorcode;
}

int tickfunc(void *parm)
{
    int len;
    char *wordptr = Word(&len);
    push((int)FindWord(wordptr, len));
    return errorcode;
}

int statefunc(void *parm)
{
    push((int)&compileflag);
    return errorcode;
}

int abortfunc(void *parm)
{
    stkidx = 0;
    retidx = 0;
    ctlidx = 0;
    return RETURN_ERROR;
}

int quitfunc(void *parm)
{
    retidx = 0;
    ctlidx = 0;
    return RETURN_QUIT;
}

int exitfunc(void *parm)
{
    return RETURN_EXIT;
}

int parenfunc(void *parm)
{
    comment = 1;
    return errorcode;
}

int bodyfunc(void *parm)
{
    dictentryT *entry = (dictentryT *)pop();
    push((int)ALIGN(entry->name + entry->len));
    return errorcode;
}

int infunc(void *parm)
{
    push((int)&inputidx);
    return errorcode;
}

int charfunc(void *parm)
{
    int len;
    char *wordptr = Word(&len);
    push((int)*wordptr);
    return errorcode;
}

int wordfunc(void *parm)
{
    int len;
    char *wordptr = Word(&len);
    pad[0] = len;
    memcpy(pad+1, wordptr, len);
    push((int)pad);
    return errorcode;
}

int executefunc(void *parm)
{
    dictentryT *entry = (dictentryT *)pop();
    if (errorcode == RETURN_ERROR) return errorcode;
    return EXECUTE(entry);
}

int dotquotefunc(void *parm)
{
    if (compileflag)
    {
        CompileString(parm);
    }
    else
    {
        char *ptr = (char *)(++execaddr);
        int num = *ptr++;
        Type(ptr, num);
        execaddr += num/4;
    }
    return errorcode;
}

int abortquotefunc(void *parm)
{
    if (compileflag)
    {
        CompileString(parm);
    }
    else
    {
        char *ptr = (char *)(++execaddr);
        int num = *ptr++;
        if (pop())
        {
            Type(ptr, num);
            stkidx = 0;
            retidx = 0;
            ctlidx = 0;
            return RETURN_ERROR;
        }
        execaddr += num/4;
    }
    return errorcode;
}

int squotefunc(void *parm)
{
    if (compileflag)
    {
        CompileString(parm);
    }
    else
    {
        char *ptr = (char *)(++execaddr);
        int num = *ptr++;
        push((int)ptr);
        push(num);
        execaddr += num/4;
    }
    return errorcode;
}

int acceptfunc(void *parm)
{
    int num = pop();
    char *ptr = (char *)pop();
    if (errorcode == RETURN_ERROR) return errorcode;
    if (fgets(ptr, num, stdin) > 0)
    {
        RemoveCRLF(ptr);
        push(strlen(ptr));
    }
    else
        push(0);
    return errorcode;
}

int keyfunc(void *parm)
{
    int value;
    while ((value = getch()) == -1);
    push(value);
    return errorcode;
}

int findfunc(void *parm)
{
    char *ptr = (char *)pop();
    int num = *ptr;
    dictentryT *entry = FindWord(ptr+1, num);
    if (!entry)
    {
        push((int)ptr);
        push(0);
        return errorcode;
    }
    push((int)entry);
    if (entry->flags & FLAG_IMMEDIATE)
        push(1);
    else
        push(-1);
    return errorcode;
}

int digit2ascii(int digit)
{
    if (digit < 10)
        digit += '0';
    else
        digit += 'A' - 10;
    return digit;
}

static int ConvertUnsignedNumber(unsigned int value)
{
    printptr = pad + 255;
    *printptr-- = digit2ascii(value % base);
    while ((value /= base))
        *printptr-- = digit2ascii(value % base);
    return (int)(pad + 255 - printptr);
}

static int ConvertSignedNumber(int value)
{
    if (value >= 0)
        return ConvertUnsignedNumber((unsigned int)value);
    ConvertUnsignedNumber((unsigned int)(-value));
    *printptr-- = '-';
    return (int)(pad + 255 - printptr);
}

#if 1
int lessnumberfunc(void *parm)
{
    printptr = pad + 255;
    return errorcode;
}

int numberfunc(void *parm)
{
    int64_t value = pop2();
    push2(value/base);
    *printptr-- = digit2ascii((int)(value % base));
    return errorcode;
}

int numbersfunc(void *parm)
{
    uint64_t value = (uint64_t)pop2();

    *printptr-- = digit2ascii((int)(value % base));
    while ((value /= base))
    {
        *printptr-- = digit2ascii((int)(value % base));
    }
    push2(0);
    return errorcode;
}

int numbergreaterfunc(void *parm)
{
    pop2();
    push((int)(printptr + 1));
    push((int)(pad + 255 - printptr));
    return errorcode;
}

int holdfunc(void *parm)
{
    *printptr-- = pop();
    return errorcode;
}

int signfunc(void *parm)
{
    if (pop())
        *printptr-- = '-';
    return errorcode;
}
#endif

int starslashfunc(void *parm)
{
    int64_t value3 = (int64_t)pop();
    int64_t value2 = (int64_t)pop();
    int64_t value1 = (int64_t)pop();
    push((int)((value1 * value2) / value3));
    return errorcode;
}

int starslashmodfunc(void *parm)
{
    int64_t value3 = (int64_t)pop();
    int64_t value2 = (int64_t)pop();
    int64_t value1 = (int64_t)pop();
    value1 *= value2;
    push((int)(value1 % value3));
    push((int)(value1 / value3));
    return errorcode;
}

int greaternumberfunc(void *parm)
{
    int digit;
    int num = pop();
    char *ptr = (char *)pop();
    int64_t value = pop2();
    while (num > 0)
    {
        if (*ptr >= '0' && *ptr <= '9')
            digit = *ptr - '0';
        else if (*ptr >= 'A' && *ptr <= 'Z')
            digit = *ptr - 'A' + 10;
        else if (*ptr >= 'a' && *ptr <= 'z')
            digit = *ptr - 'a' + 10;
        else
            break;
        if (digit >= base) break;
        value = (value * base) + digit;
        ptr++;
        num--;
    }
    push2(value);
    push((int)ptr);
    push(num);
    return errorcode;
}

int fmslashmodfunc(void *parm)
{
    int64_t value2 = (int64_t)pop();
    int64_t value1 = pop2();
    int64_t quotient, remainder, round;
    int sign2 = (value2 < 0);
    int sign1 = (value1 < 0);

    if (sign1 == sign2)
    {
        quotient = value1 / value2;
        remainder = value1 % value2;
    }
    else
    {
        round = (value1 << 1) - 1;
        quotient = (value1 - value2 + round) / value2;
        remainder = value1 - (quotient * value2);
    }
    push((int)remainder);
    push((int)quotient);
    return errorcode;
}

int smslashremfunc(void *parm)
{
    int64_t value2 = (int64_t)pop();
    int64_t value1 = pop2();
    push((int)(value1 % value2));
    push((int)(value1 / value2));
    return errorcode;
}

int mstarfunc(void *parm)
{
    int64_t value2 = (int64_t)pop();
    int64_t value1 = (int64_t)pop();
    push2(value1 * value2);
    return errorcode;
}

int mstarslashfunc(void *parm)
{
    int64_t value3 = (int64_t)pop();
    int64_t value2 = (int64_t)pop();
    int64_t value1 = pop2();
    push2((value1 * value2) / value3);
    return errorcode;
}

int unsignedmstarfunc(void *parm)
{
    uint64_t value2 = (uint64_t)pop();
    uint64_t value1 = (uint64_t)pop();
    push2(value1 * value2);
    return errorcode;
}

int unsignedmslashmodfunc(void *parm)
{
    uint64_t value2 = (uint64_t)pop();
    uint64_t value1 = (uint64_t)pop();
    push((int)(value1 % value2));
    push((int)(value1 / value2));
    return errorcode;
}

int bracketcharfunc(void *parm)
{
    int len;
    char *wordptr;

    if (!compileflag)
        return RETURN_ERROR;

    wordptr = Word(&len);

    if (!len)
        return RETURN_ERROR;

    here = ALIGN(here);
    *((int *)here) = (int)literalword;
    here += 4;
    *((int *)here) = *wordptr;
    here += 4;

    return errorcode;
}

int brackettickfunc(void *parm)
{
    int len;
    char *wordptr;
    dictentryT *wordentry = 0;

    if (!compileflag)
        return RETURN_ERROR;

    wordptr = Word(&len);
    wordentry = FindWord(wordptr, len);

    if (!wordentry)
        return RETURN_ERROR;

    here = ALIGN(here);
    *((int *)here) = (int)literalword;
    here += 4;
    *((int *)here) = (int)wordentry;
    here += 4;

    return errorcode;
}

int FAST tofunc(void *parm)
{
    int len;
    char *wordptr = Word(&len);
    dictentryT *wordentry = FindWord(wordptr, len);

    if (!wordentry)
    {
        Type(wordptr, len);
        printf(" ?\n");
        return RETURN_ERROR;
    }

    if (compileflag)
    {
        here = ALIGN(here);
        *((int *)here) = (int)literalword;
        here += 4;
        *((int *)here) = (int)BODY(wordentry);
        here += 4;
        *((int *)here) = (int)storeword;
        here += 4;
    }
    else
    {
        int *ptr = (int *)BODY(wordentry);
        if (stkidx < 1)
            return RETURN_ERROR;
        *ptr = stack[--stkidx];
    }

    return errorcode;
}

int wordsfunc(void *parm)
{
    int len = 0;
    dictentryT *wordentry = lastword;

    while (wordentry)
    {
        if (len + wordentry->len > 78)
        {
            printf("\n");
            len = wordentry->len;
        }
        else if (len)
        {
            printf(" ");
            len += wordentry->len + 1;
        }
        else
            len = wordentry->len;
        Type(wordentry->name, wordentry->len);
        wordentry = wordentry->prev;
    }
    printf("\n");
    return errorcode;
}

int seefunc(void *parm)
{
    int len;
    char *wordptr = Word(&len);
    dictentryT *wordentry = FindWord(wordptr, len);

    if (!wordentry)
    {
        Type(wordptr, len);
        printf(" ?\n");
        return RETURN_ERROR;
    }

    Type(wordentry->name, wordentry->len);
    if (wordentry->func == execfunc)
    {
        dictentryT *entry;
        parm = ALIGN(wordentry->name + wordentry->len);
        while (*(int *)parm)
        {
            entry = (dictentryT *)(*(int *)parm);
#ifndef __PROPELLER__
            if (((int)entry) & 0xff000000 || ((int)entry) < 10000)
            {
                printf("error: entry = %d\n", (int)entry);
                parm += 4;
                continue;
            }
#endif
            printf(" ");
            Type(entry->name, entry->len);
            parm += 4;
            if (entry->flags & FLAG_LITERAL)
            {
                int len = ConvertSignedNumber(*(int *)parm);
                Type(printptr+1, len);
                parm += 4;
            }
            else if (entry->flags & FLAG_STRING)
            {
                int num = *(char *)parm++;
                Type(parm, num);
                parm += num;
                parm = ALIGN(parm);
            }
        }
    }
    else if (wordentry->func == varexecfunc)
    {
        int len;
        int *ptr = (int *)ALIGN(wordentry->name + wordentry->len);
        len = ConvertSignedNumber(ptr[0]);
        Type(printptr+1, len);
        len = ConvertSignedNumber(ptr[1]);
        Type(printptr+1, len);
    }
    printf("\n");

    return errorcode;
}

int forgetfunc(void *parm)
{
    int len;
    char *wordptr = Word(&len);
    dictentryT *wordentry = FindWord(wordptr, len);

    if (!wordentry)
    {
        Type(wordptr, len);
        printf(" ?\n");
        return RETURN_ERROR;
    }

    here = (char *)wordentry;
    lastword = wordentry->prev;

    return errorcode;
}

int dotsfunc(void *parm)
{
    int i, len;
    printf("<");
    len = ConvertSignedNumber(stkidx);
    Type(printptr+1, len);
    printf("> ");
    for (i = 0; i < stkidx; i++)
    {
        len = ConvertSignedNumber(stack[i]);
        Type(printptr+1, len);
        printf(" ");
    }
    printf("\n");
    return errorcode;
}

int unusedfunc(void *parm)
{
    push((int)(((char *)&codespace[CODE_SIZE]) - here));
    return errorcode;
}

#if 1
int hexfunc(void *parm)
{
    base = 16;
    return errorcode;
}

int decimalfunc(void *parm)
{
    base = 10;
    return errorcode;
}

int FAST ifunc(void *parm)
{
    stack[stkidx++] = retstk[retidx-1];
    return errorcode;
}

int FAST jfunc(void *parm)
{
    stack[stkidx++] = retstk[retidx-3];
    return errorcode;
}
#endif

int twodupfunc(void *parm)
{
    int value1 = pop();
    int value2 = pop();
    push(value2);
    push(value1);
    push(value2);
    push(value1);
    return errorcode;
}

int overfunc(void *parm)
{
    int value1 = pop();
    int value2 = pop();
    push(value2);
    push(value1);
    push(value2);
    return errorcode;
}

int FAST modfunc(void *parm)
{
    int value1, value2;
    if (stkidx < 2) return RETURN_ERROR;
    value2 = stack[--stkidx];
    value1 = stack[stkidx-1];
    stack[stkidx-1] = value1 % value2;
    return errorcode;
}

int movefunc(void *parm)
{
    int num = pop();
    int addr2 = pop();
    int addr1 = pop();
    memmove((char *)addr2, (char *)addr1, num);
    return errorcode;
}

int withinfunc(void *parm)
{
    int value3 = pop();
    int value2 = pop();
    int value1 = pop();
    if (value2 < value3)
        push(-(value2 <= value1 && value1 < value3));
    else if (value2 > value3)
        push(-(value2 <= value1 || value1 < value3));
    else
        push(0);
    return errorcode;
}

void InitializeDictionary()
{
    MakeEntry("+", plusfunc, 0);
    MakeEntry("-", minusfunc, 0);
    MakeEntry("*", multfunc, 0);
    MakeEntry("/", dividefunc, 0);
    MakeEntry(".", dotfunc, 0);
    MakeEntry("!", storefunc, 0);
    storeword = lastword;
    MakeEntry("@", fetchfunc, 0);
    MakeEntry("c!", cstorefunc, 0);
    MakeEntry("c@", cfetchfunc, 0);
    MakeEntry("here", herefunc, 0);
    MakeEntry("create", createfunc, 0);
    MakeEntry("allot", allotfunc, 0);
    MakeEntry("(literal)", _literalfunc, FLAG_LITERAL);
    literalword = lastword;
    MakeEntry("postpone", postponefunc, 0);
    postponeword = lastword;
    MakeEntry(":", colonfunc, 0);
    MakeEntry("[", compileofffunc, FLAG_IMMEDIATE);
    MakeEntry("]", compileonfunc, FLAG_IMMEDIATE);
    MakeEntry(";", semicolonfunc, FLAG_IMMEDIATE);
    MakeEntry("immediate", immediatefunc, 0);
    MakeEntry("include", includefunc, 0);
    MakeEntry(">r", toretfunc, 0);
    MakeEntry("r>", fromretfunc, 0);
    MakeEntry("rot", rotfunc, 0);
    MakeEntry("drop", dropfunc, 0);
    MakeEntry("dup", dupfunc, 0);
    MakeEntry("swap", swapfunc, 0);
    MakeEntry("=", equalfunc, 0);
    MakeEntry("<", lessfunc, 0);
    MakeEntry(">", greaterfunc, 0);
    MakeEntry("emit", emitfunc, 0);
    MakeEntry("and", andfunc, 0);
    MakeEntry("or", orfunc, 0);
    MakeEntry("xor", xorfunc, 0);
    MakeEntry("depth", depthfunc, 0);
    MakeEntry("begin", beginfunc, FLAG_IMMEDIATE);
    MakeEntry("until", untilfunc, FLAG_IMMEDIATE | FLAG_LITERAL);
    MakeEntry("again", againfunc, FLAG_IMMEDIATE | FLAG_LITERAL);
    MakeEntry("do", dofunc, FLAG_IMMEDIATE);
    MakeEntry("loop", loopfunc, FLAG_IMMEDIATE | FLAG_LITERAL);
    MakeEntry("+loop", plusloopfunc, FLAG_IMMEDIATE | FLAG_LITERAL);
    MakeEntry("if", iffunc, FLAG_IMMEDIATE | FLAG_LITERAL);
    MakeEntry("else", elsefunc, FLAG_IMMEDIATE | FLAG_LITERAL);
    MakeEntry("then", thenfunc, FLAG_IMMEDIATE);
    MakeEntry(".\"", dotquotefunc, FLAG_IMMEDIATE | FLAG_STRING);
    MakeEntry("bye", byefunc, 0);
    MakeEntry("rshift", rshiftfunc, 0);
    MakeEntry("lshift", lshiftfunc, 0);
    MakeEntry("base", basefunc, 0);
    MakeEntry("u.", udotfunc, 0);
    MakeEntry("u<", ulessfunc, 0);
    MakeEntry("type", typefunc, 0);
    MakeEntry("source", sourcefunc, 0);
    MakeEntry("state", statefunc, 0);
    MakeEntry("'", tickfunc, 0);
    MakeEntry("abort", abortfunc, 0);
    MakeEntry("abort\"", abortquotefunc, FLAG_IMMEDIATE | FLAG_STRING);
    MakeEntry("quit", quitfunc, 0);
    MakeEntry("exit", exitfunc, 0);
    MakeEntry("(", parenfunc, FLAG_IMMEDIATE);
    MakeEntry(">BODY", bodyfunc, 0);
    MakeEntry(">IN", infunc, 0);
    MakeEntry("char", charfunc, 0);
    MakeEntry("word", wordfunc, 0);
    MakeEntry("execute", executefunc, 0);
    MakeEntry("literal", literalfunc, FLAG_IMMEDIATE);
    MakeEntry("s\"", squotefunc, FLAG_IMMEDIATE | FLAG_STRING);
    MakeEntry("does>", doesfunc, 0);
    MakeEntry("while", whilefunc, FLAG_IMMEDIATE | FLAG_LITERAL);
    MakeEntry("repeat", repeatfunc, FLAG_IMMEDIATE | FLAG_LITERAL);
    MakeEntry("accept", acceptfunc, 0);
    MakeEntry("key", keyfunc, 0);
    MakeEntry("find", findfunc, 0);
    MakeEntry("recurse", recursefunc, FLAG_IMMEDIATE);
    MakeEntry("evaluate", evaluatefunc, 0);
#if 1
    MakeEntry("<#", lessnumberfunc, 0);
    MakeEntry("#", numberfunc, 0);
    MakeEntry("#>", numbergreaterfunc, 0);
    MakeEntry("#s", numbersfunc, 0);
    MakeEntry("hold", holdfunc, 0);
    MakeEntry("sign", signfunc, 0);
#endif
    MakeEntry("*/", starslashfunc, 0);
    MakeEntry("*/mod", starslashmodfunc, 0);
    MakeEntry(">number", greaternumberfunc, 0);
    MakeEntry("fm/mod", fmslashmodfunc, 0);
    MakeEntry("sm/rem", smslashremfunc, 0);
    MakeEntry("m*", mstarfunc, 0);
    MakeEntry("m*/", mstarslashfunc, 0);
    MakeEntry("um*", unsignedmstarfunc, 0);
    MakeEntry("um/mod", unsignedmslashmodfunc, 0);
    MakeEntry("[char]", bracketcharfunc, FLAG_IMMEDIATE);
    MakeEntry("[']", brackettickfunc, FLAG_IMMEDIATE);
    MakeEntry(".r", dotrfunc, 0);
    MakeEntry("leave", leavefunc, FLAG_IMMEDIATE | FLAG_LITERAL);
    MakeEntry("?do", questiondofunc, FLAG_IMMEDIATE | FLAG_LITERAL);
    MakeEntry("to", tofunc, FLAG_IMMEDIATE);
    MakeEntry("see", seefunc, 0);
    MakeEntry("words", wordsfunc, 0);
    MakeEntry("forget", forgetfunc, 0);
    MakeEntry(".s", dotsfunc, 0);
    MakeEntry("unused", unusedfunc, 0);
#if 1
    MakeEntry("decimal", decimalfunc, 0);
    MakeEntry("hex", hexfunc, 0);
    MakeEntry("i", ifunc, 0);
    MakeEntry("j", jfunc, 0);
#endif
    MakeEntry("2dup", twodupfunc, 0);
    MakeEntry("over", overfunc, 0);
    MakeEntry("mod", modfunc, 0);
    MakeEntry("move", movefunc, 0);
    MakeEntry("within", withinfunc, 0);
}
/*
+--------------------------------------------------------------------
|  TERMS OF USE: MIT License
+--------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files
(the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge,
publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
+------------------------------------------------------------------
*/
